/* modules.c - Compiled-in module support
 * Copyright 2002-2003 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.  Important limitations are
 * listed in the COPYING file that accompanies this software.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, email srvx-maintainers@srvx.net.
 *
 * $Id: modules.c,v 1.26 2003/07/07 16:29:33 entrope Exp $
 */

#include "log.h"
#include "modules.h"
#include "modules-list.h"

enum init_state {
    UNINIT,
    WORKING,
    BROKEN,
    DONE
};

struct cmodule {
    const char *name;
    int (*init_func)(void);
    const char **deps;
    enum init_state state;
};

#ifdef WITH_MODULE_SNOOP
extern int snoop_init(void);
const char *snoop_deps[] = { NULL };
#endif

#ifdef WITH_MODULE_SERVERSPY
extern int serverspy_init(void);
const char *serverspy_deps[] = { NULL };
#endif

static struct cmodule cmodules[] = {
#ifdef WITH_MODULE_SNOOP
    { "snoop", snoop_init, snoop_deps, UNINIT },
#endif
#ifdef WITH_MODULE_SERVERSPY
    { "serverspy", serverspy_init, serverspy_deps, UNINIT },
#endif
    /* Placeholder at end of array */
    { NULL, NULL, NULL, UNINIT }
};

static int
modules_bsearch(const void *a, const void *b) {
    const char *key = a;
    const struct cmodule *cmod = b;
    return irccasecmp(key, cmod->name);
}

static int
modules_qsort(const void *a, const void *b) {
    const struct cmodule *ca = a, *cb = b;
    return irccasecmp(ca->name, cb->name);
}

static int
module_init(struct cmodule *cmod) {
    unsigned int ii;
    struct cmodule *dep;

    switch (cmod->state) {
    case UNINIT: break;
    case DONE: return 1;
    case BROKEN: return 0;
    case WORKING:
        log_module(MAIN_LOG, LOG_ERROR, "Tried to recursively enable code module %s.", cmod->name);
        return 0;
    }
    log_module(MAIN_LOG, LOG_INFO, "Starting initialization of code module %s.", cmod->name);
    cmod->state = WORKING;
    for (ii=0; cmod->deps[ii]; ++ii) {
        dep = bsearch(cmod->deps[ii], cmodules, ArrayLength(cmodules)-1, sizeof(cmodules[0]), modules_bsearch);
        if (!dep) {
            log_module(MAIN_LOG, LOG_ERROR, "Code module %s depends on unknown module %s.", cmod->name, cmod->deps[ii]);
            cmod->state = BROKEN;
            return 0;
        }
        if (!module_init(dep)) {
            log_module(MAIN_LOG, LOG_ERROR, "Failed to initialize dependency %s of code module %s.", dep->name, cmod->name);
            cmod->state = BROKEN;
            return 0;
        }
    }
    if (!cmod->init_func()) {
        log_module(MAIN_LOG, LOG_ERROR, "Failed to initialize code module %s.", cmod->name);
        cmod->state = BROKEN;
        return 0;
    }
    cmod->state = DONE;
    return 1;
}

void
modules_init(void) {
    unsigned int ii;

    qsort(cmodules, ArrayLength(cmodules)-1, sizeof(cmodules[0]), modules_qsort);
    for (ii=0; cmodules[ii].name; ++ii) {
        if (cmodules[ii].state == UNINIT) {
            module_init(cmodules + ii);
        }
        if (cmodules[ii].state != DONE) {
            log_module(MAIN_LOG, LOG_WARNING, "Code module %s not properly initialized.", cmodules[ii].name);
        }
    }
}
